{-# LANGUAGE DataKinds           #-}
{-# LANGUAGE GADTs               #-}
{-# LANGUAGE KindSignatures      #-}
{-# LANGUAGE ScopedTypeVariables #-}
{-# LANGUAGE StandaloneDeriving  #-}
module Data.FixedWidth.Int (FixedInt) where

import           Data.Bifunctor
import           Data.Bits
import           Data.FixedWidth.Shared
import           Data.Int
import           Data.Kind
import           Data.Proxy
import           Data.Ratio
import           GHC.TypeLits

data FixedInt :: Nat -> Type where
    FixedInt :: KnownNat n => Int64 -> FixedInt n

deriving instance Eq (FixedInt n)
deriving instance Ord (FixedInt n)

instance Show (FixedInt n) where
    show (FixedInt n) = show n

instance (KnownNat n) => Bounded (FixedInt n) where
    minBound = FixedInt $ -(2 ^ (width - 1))
        where width = bitWidth (FixedInt 0 :: FixedInt n)
    maxBound = FixedInt $ 2 ^ (width - 1) - 1
        where width = bitWidth (FixedInt 0 :: FixedInt n)

instance (KnownNat n) => Enum (FixedInt n) where
    toEnum v
        | FixedInt v >= (minBound :: FixedInt n) && FixedInt v <= (maxBound :: FixedInt n) = FixedInt v
        | otherwise = error $ concat ["\"", show v, "\" is too large to fit in an ", show $ bitWidth (FixedInt 0 :: FixedInt n), "-bit integer"]
    fromEnum (FixedInt v) = v

instance (KnownNat n) => Num (FixedInt n) where
    (FixedInt a) + (FixedInt b) = FixedInt $ if res > max then res .&. max - (max + 1) else res
        where
            res = a + b
            (FixedInt min) = minBound :: FixedInt n
            (FixedInt max) = maxBound :: FixedInt n
    (FixedInt a) * (FixedInt b) = FixedInt $ if res > max || res < min then res .&. max - (max + 1) else res
        where
            res = a * b
            (FixedInt min) = minBound :: FixedInt n
            (FixedInt max) = maxBound :: FixedInt n
    (FixedInt a) - (FixedInt b) = FixedInt $ if res < min then res .&. max - (max + 1) else res
        where
            res = a - b
            (FixedInt min) = minBound :: FixedInt n
            (FixedInt max) = maxBound :: FixedInt n
    negate f@(FixedInt a) = if f == minBound then minBound else FixedInt $ negate a
    abs f@(FixedInt a) = if f == minBound then minBound else FixedInt $ abs a
    signum (FixedInt a) = FixedInt $ signum a
    fromInteger int
        | int >= min' && int <= max' = FixedInt $ fromInteger int
        | otherwise = error $ concat ["\"", show int, "\" is too large to fit in an ", show $ bitWidth (FixedInt 0 :: FixedInt n), "-bit integer"]
        where
            (FixedInt min) = minBound :: FixedInt n
            (FixedInt max) = maxBound :: FixedInt n
            min' = toInteger min
            max' = toInteger max

instance (KnownNat n) => Real (FixedInt n) where
    toRational (FixedInt a) = (toInteger a) % 1

instance (KnownNat n) => Integral (FixedInt n) where
    (FixedInt a) `quotRem` (FixedInt b) = bimap FixedInt FixedInt $ quotRem a b
    toInteger (FixedInt a) = toInteger a
